(*GJMAT.PAS - Gauss-Jordan matrix elimination

  Title   : GJMAT
  Version : 1.3
  Date    : March 19, 2000
  Author  : J R Ferguson
  Compiler: Borland Pascal v4.0 or up
  Usage   : Menu oriented DOS program
  Download: http://hello.to/ferguson
  E-mail  : j.r.ferguson@iname.com

**English**
  Gauss-Jordan matrix elimination.
  Solution of a set of linear equations.
  Also computation of determinant and inverse of a square matrix.

  Modeled after a Basic program by Patrick E. McGuire
  from BYTE, August 1983, page 394.

  This program and its source may be used and copied freely without charge,
  but only for non-commercial purposes. The author is not responsible for any
  damage or loss of data that may be caused by using it.

  To compile this source file, you will need unit ConLib from my JRFPAS Pascal
  routine library, which can be downloaded from the Internet website mentioned
  above.

  To create a Dutch version of this program, you may have to edit the language
  switch below this comment block first, changing:
    {$define Lang_US}  {US English}
    {$undef  Lang_NL}  {Dutch}
  into:
    {$undef  Lang_US}  {US English}
    {$define Lang_NL}  {Dutch}
  After compiling the program you can change the name of the executable file
  if you like.


**Nederlands (Dutch)**
  Gauss-Jordan matrix eliminatie.
  Oplossing van een stelsel lineaire vergelijkingen.
  Tevens berekening van de determinant en de inverse van een vierkante matrix.

  Naar een Basic programma door Patrick E. McGuire
  uit BYTE, augustus 1983, pag. 394.

  Dit programma en zijn broncode mag zonder vergoeding worden gekopieerd en
  gebruikt, maar uitsluitend zonder winstoogmerk. De auteur kan op geen
  enkele manier aansprakelijk worden gesteld voor vergoeding van schade of
  verlies van gegevens dat door het gebruik ervan zou kunnen zijn
  veroorzaakt.

  Om het programma te kunnen compileren heeft u unit ConLib nodig uit mijn
  JRFPAS Pascal routinebibliotheek. Deze kan gedownload worden van de
  bovengenoemde internet website.

  Om een Nederlandstalige versie te maken moet men zonodig eerste de taalkeuze
  aanpassen door de regels:
    {$define Lang_US}  {US English}
    {$undef  Lang_NL}  {Dutch}
  onder dit commentaarblok te modify in:
    {$undef  Lang_US}  {US English}
    {$define Lang_NL}  {Dutch}
  Na het compileren kunt u de naam van het aangemaakte programmabestand
  desgewenst modify.
*)

{Language switch : define exactly one}
{$define Lang_US}  {US English}
{$undef  Lang_NL}  {Dutch}


program gjmat(input,output,data);
uses
{$ifdef Lang_US} gjmatus, {$endif}
{$ifdef Lang_NL} gjmatnl, {$endif}
ConLib, Crt;

const
  maxrow       = 20;
  maxcol       = 41;        { = 2*maxrow + 1 }
  dflfnm       = 'TMP';     { default datafile name }
  dflext       = '.DAT';    { default datafile extension }
  dflfw        = 10;        { default field width }
  dflnd        = 4;         { default number of digits }

type
  rowindex     = 1..maxrow;
  colindex     = 1..maxcol;
  vector       = array[rowindex] of real;
  matrix       = array[rowindex,colindex] of real;
  filename     = string[14];

var
  n            : rowindex; { actual number of rows }
  a            : matrix;
  det          : real;
  option       : char;
  hasinput,
  computed     : boolean;
  data         : text;
  datafile     : filename;
  fw,nd        : integer;

function betweenint(l,i,h: integer): boolean;
  begin betweenint:= (l <= i) and (i <= h) end;

procedure beep; begin write(chr(7)) end;

procedure errbeep(ok: boolean);
begin if not ok then beep end;

procedure error; begin beep; WaitCr end;

procedure readint(var f:text; var i:integer; var ok: boolean);
  begin
    {$I-} readln(f,i); {$I+}
    ok:= (ioresult = 0)
  end;

procedure readreal(var f: text; var r:real; var ok: boolean);
  begin
    {$I-} readln(f,r); {$I+}
    ok:= ioresult = 0
  end;

function exists(var name: filename): boolean;
  var f: file;
  begin
   Assign(f,name);
   {$I-} reset(f) {$I+};
   exists:= IOresult = 0
  end;

procedure openread(var f: text; var name: filename);
  begin
    Assign(f,name); reset(f)
  end;

procedure openwrite(var f: text; var name: filename);
  begin
    Assign(f,name); rewrite(f)
  end;

procedure closeread (var f: text); begin Close(f) end;

procedure closewrite(var f: text); begin Close(f) end;


procedure initglobals;
  begin
    hasinput := false;
    computed := false;
    datafile := dflfnm+dflext;
    fw       := dflfw;
    nd       := dflnd
  end;

procedure displaymatrix(var a: matrix; n: rowindex;
                        startcol, endcol: colindex);
  var r : rowindex;
      k : colindex;
  begin
    for r:= 1 to n do
      begin
        write(r:2,': ');
        for k:= startcol to endcol do
          if nd = -1 then
            begin
              if a[r,k] < 0 then write('-') else write(' ');
              write(abs(a[r,k]):fw-1,' ')
            end
          else write(a[r,k]:fw:nd,' ');
        writeln
      end;
    writeln
  end;

procedure gaussjordan(n: rowindex; var a: matrix;
                      var det: real; var hassolution: boolean);
  var r,s,t : rowindex;
      h,j,k : colindex;
      d,q   : real;
  begin

    { initialize the inverse as an identity matrix}

    j:= n+1;
    for r:= 1 to n do
      begin
        j:= j+1;
        for k:= n+2 to 2*n+1 do if k=j then a[r,k]:= 1.0 else a[r,k]:= 0.0
      end;

    { solve }

    det:= 1.0; hassolution:= true;
    for r:= 1 to n do if hassolution then
      begin
        { -- pivot row normalization -- }
        j:= 2*n+1;
        d:= a[r,r]; if d = 0.0 then hassolution:= false;
        if hassolution then
          begin
            det:= det * d;
            for h:= 1 to j do a[r,h]:= a[r,h] / d;
            { -- non-pivot row reduction -- }
            for s:= 1 to n do
              begin
                q:= a[s,r];
                if s<>r then for t:= 1 to j do a[s,t]:= a[s,t] - q * a[r,t]
              end
          end
      end

  end;

procedure errormessage;
  begin
         if not hasinput then begin writeln('> '+S_NoData)    ; error end
    else if not computed then begin writeln('> '+S_NoSolution); error end
  end;

procedure solve(var n: rowindex; var a: matrix; var det: real);
  begin
    if not computed then gaussjordan(n,a,det,computed);
    errormessage
  end;

procedure determinant(var n: rowindex; var a:matrix; var det: real);
  begin
    solve(n,a,det);
    if computed then begin writeln (S_Determinant+' = ',det); WaitCr end;
  end;

procedure inverse(var n: rowindex; var a:matrix);
  begin
    solve(n,a,det);
    if computed then
      begin
        ClrScr; writeln(S_Inverse+':'); writeln;
        displaymatrix(a,n,n+2,2*n+1);
        WaitCr
      end
  end;

procedure solution(var n: rowindex; var a:matrix);
  begin
    solve(n,a,det);
    if computed then
      begin
        ClrScr; writeln(S_Solution+':'); writeln;
        solve(n,a,det);
        displaymatrix(a,n,n+1,n+1);
        WaitCr
      end
  end;

procedure fieldwidth(var fw: integer);
  var ok: boolean;
  begin
    ClrScr;
    writeln(S_FieldWHeader);
    writeln;
    repeat
      write(S_FieldWPrompt+' <1..99>  [',fw:2,'] : ');
      fw:= dflfw;
      readint(input,fw,ok); ok:= ok and betweenint(1,fw,99); errbeep(ok)
    until ok
  end;

procedure digits(var nd: integer);
  var ok: boolean;
  begin
    ClrScr;
    writeln(S_DigitsHeader);
    writeln('-1 : '+S_FloatingPoint);
    writeln;
    repeat
      write(S_DigitsPrompt+' <-1..99>  [',nd:2,'] : ');
      nd:= dflnd;
      readint(input,nd,ok); ok:= ok and betweenint(-1,nd,99); errbeep(ok)
    until ok
  end;

procedure askname(var name: filename);
  var oldname : filename;
      i       : integer;
      ext     : boolean;
  begin
    oldname:= datafile;
    ClrScr; write(S_FileNmHeader+'  [',datafile,'] : '); readln(name);
    ext:= false;
    for i:= 1 to length(name) do
      if name[i] = '.' then ext:= true else name[i]:= UpCase(name[i]);
    if name = '' then name:= oldname
    else if not ext then name:= name+dflext
  end;

procedure readfile(var n: rowindex; var a:matrix);
  var number : integer;
      r      : rowindex;
      k      : colindex;
      ok     : boolean;

  begin
    if exists(datafile) then
      begin
        write(S_Reading+' ... ');
        openread(data,datafile);
        readint(data,number,ok); ok:= ok and betweenint(2,number,maxrow);
        if ok then
          begin
            n:= number;
            for r:= 1 to n do for k:= 1 to n do
              if ok then readreal(data,a[r,k],ok);
            for r:= 1 to n do
              if ok then readreal(data,a[r,n+1],ok)
          end;
        closeread(data);
        hasinput:= ok;
        computed:= false;
        if not ok then
          begin writeln('> '+S_ErrData); error end;
      end
    else begin writeln('> '+S_ErrNotFound); error end
  end;

procedure manual(var n: rowindex; var a:matrix);
  var number : integer;
      r      : rowindex;
      k      : colindex;
      ok     : boolean;
  begin
    ClrScr; writeln(S_InputHeader+':'); writeln;
    ok:= true;
    repeat
      write(S_InputRows+' <2..',maxrow:2,'>  : ');
      readint(input,number,ok);
      ok:= ok and betweenint(2,number,maxrow); errbeep(ok)
    until ok;
    n:= number;
    for r:= 1 to n do
      begin
        writeln;
        writeln('row ',r:2,':');
        for k:= 1 to n do
          repeat
            write('A(',r:2,',',k:2,') = ');
            readreal(input,a[r,k],ok); errbeep(ok)
          until ok;
        repeat
          write('B(',r:2,')    = ');
          readreal(input,a[r,n+1],ok); errbeep(ok)
        until ok
      end;
    hasinput:= true; computed:= false
  end;

procedure modify(var n: rowindex; var a:matrix);
  var row,col  : integer;
      ok       : boolean;
      oldvalue : real;
  begin
    ok:= true;
    if hasinput then
      repeat
        ClrScr; writeln(S_CheckData); writeln;
        displaymatrix(a,n,1,n+1);
        repeat
          write(S_Row+' <1..',n:2,'> '+S_OrStop+'   : ');
          row:= 0;
          readint(input,row,ok); ok:= ok and betweenint(0,row,n); errbeep(ok)
        until ok;
        if row>0 then
          begin
            repeat
              write(S_Column+' <1..',n+1:2,'>               : ');
              col:= 0;
              readint(input,col,ok);
              ok:= ok and betweenint(1,col,n+1); errbeep(ok)
            until ok;
            oldvalue:= a[row,col];
            repeat;
              write ('A[',row:2,',',col:2,'] = ');
              readreal(input,a[row,col],ok); errbeep(ok)
            until ok;
            if a[row,col] <> oldvalue then computed:= false
          end
      until row = 0
    else errormessage
  end;

procedure writefile(var n: rowindex; var a:matrix);
  var r: rowindex;
      k: colindex;
  begin
    if hasinput then
      begin
        write(S_Writing+' ...');
        openwrite(data,datafile);
        writeln(data,n);
        for r:= 1 to n do for k:= 1 to n do writeln(data,a[r,k]);
        for r:= 1 to n do writeln(data,a[r,n+1]);
        closewrite(data)
      end
    else errormessage
  end;

begin {program}
initglobals;
repeat
  ClrScr; writeln(S_ProgTitle);
  writeln;
  writeln(S_MnuInput);
  writeln('  '+S_MnuIManual);
  writeln('  '+S_MnuICheck);
  writeln('  '+S_MnuIRead);
  writeln('  '+S_MnuIWrite);
  writeln;
  writeln(S_MnuCompute);
  writeln('  '+S_MnuCDeterm);
  writeln('  '+S_MnuCInvert);
  writeln('  '+S_MnuCSolve);
  writeln;
  writeln(S_MnuFormat);
  writeln('  '+S_MnuFFWidth+' [',fw:2,']');
  writeln('  '+S_MnuFDigits+' [',nd:2,']');
  writeln;
  writeln(S_MnuGeneral);
  writeln('  '+S_MnuGFileNm+' [',datafile,']');
  writeln('  '+S_MnuGExit);
  writeln;
  write  (S_Selection); option:= UppKey; writeln(option);
  writeln;

  if option in
    [mc_IManual,
     mc_ICheck,
     mc_IRead,
     mc_IWrite,
     mc_CDeterm,
     mc_CInvert,
     mc_CSolve,
     mc_FFWidth,
     mc_FDigits,
     mc_GFileNm,
     mc_GExit]
  then case option of

    mc_IManual : manual(n,a);
    mc_ICheck  : modify(n,a);
    mc_IRead   : readfile(n,a);
    mc_IWrite  : writefile(n,a);

    mc_CDeterm : determinant(n,a,det);
    mc_CInvert : inverse(n,a);
    mc_CSolve  : solution(n,a);

    mc_FFWidth : fieldwidth(fw);
    mc_FDigits : digits(nd);

    mc_GFileNm : askname(datafile);
    mc_GExit   : writeln(S_Exit);

  end;

until option=mc_GExit
end.
